/*
 *	shm.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

#include "hardware.h"

XImage *createShmImage( Display *dpy, int width, int height,
   XShmSegmentInfo *info )
{
	XImage *image;
	int major,minor,pixmap;

	/* set values for undo in case of error */
	image=NULL;
	info->shmid=-1;
	info->shmaddr=(char*)-1;

	/* check if X server support XShm extension */
	if (!XShmQueryVersion(dpy, &major, &minor, &pixmap)){
	   fprintf(stderr,
	      "XShm extension not supported. Use non-XShm XImage instead.\n");
	   UseXShm=FALSE;
	}

	if (UseXShm){
	   image = XShmCreateImage( dpy,
	      DefaultVisual(dpy,DefaultScreen(dpy)),
	      DefaultDepth(dpy,DefaultScreen(dpy)), ZPixmap, NULL,
	         info, width, height );
	   if (!image) {
	      fprintf( stderr,
	         "XShmCreateImage failed. Fall back to non-XShm XImage.\n" );
	      goto no_Xshm;
	   }

	   info->shmid = shmget( IPC_PRIVATE, image->bytes_per_line *
	      image->height, IPC_CREAT|0777 );
	   if (info->shmid==-1) {
	      fprintf( stderr, "shmget failed\n" );
	      goto error;
	   }

	   info->shmaddr = image->data = (char*)shmat( info->shmid, NULL, 0 );
	   info->readOnly = False;
	   if ( !XShmAttach( dpy, info ) ) {
	      fprintf( stderr, "XShmAttach failed\n" );
	      goto error;
	   }
	   return image;
	}

no_Xshm:

	image = XCreateImage( dpy,
	   DefaultVisual(dpy,DefaultScreen(dpy)),
	   DefaultDepth(dpy,DefaultScreen(dpy)), ZPixmap, 0, NULL,
	   width, height, 32, 0 );

	if (!image) {
	   fprintf( stderr, "XCreateImage failed\n" );
	   goto error;
	}
	image->data=malloc(image->height*image->bytes_per_line);
	return image;

error:
	/* undo what have done so far */

	if (UseXShm){
	   if (info->shmaddr!=(char*)-1)
	      shmdt( info->shmaddr );
	   if (info->shmid!=-1)
	      shmctl( info->shmid, IPC_RMID, NULL );
	}
	if(image!=NULL)
	   XDestroyImage(image);
	return NULL;
}

void destroyShmImage(Display *dpy, XImage *image, XShmSegmentInfo *info)
{
	if (UseXShm) {
	   XShmDetach( dpy, info );
	   XDestroyImage( image );
	   if (shmctl( info->shmid, IPC_RMID, NULL )==-1){
	      perror("destroyShmImage() cannot remove share memory");
	   }
	}
	else
	   XDestroyImage( image );
}
